"""RabbitMQ Executor."""

import re
import subprocess
from pathlib import Path
from typing import List, Optional

from mirakuru import TCPExecutor


class RabbitMqExecutor(TCPExecutor):
    """RabbitMQ executor to start specific rabbitmq instances."""

    _UNWANTED_QUEUE_PATTERN = re.compile("(done|timeout:|listing queues)")

    def __init__(
        self,
        command: str,
        host: str,
        port: int,
        distribution_port: int,
        rabbit_ctl: str,
        logpath: Path,
        path: Path,
        plugin_path: Path,
        node_name: Optional[str] = None,
    ) -> None:  # pylint:disable=too-many-arguments
        """Initialize RabbitMQ executor.

        :param command: rabbitmq-server location
        :param host: host where rabbitmq will be accessible
        :param port: port under which rabbitmq runs
        :param rabbit_ctl: rabbitctl location
        :param logpath:
        :param path: Path containing rabbitmq'a mnesia na plugins
        :param node_name: RabbitMQ node name
        """
        envvars = {
            "RABBITMQ_LOG_BASE": str(logpath / f"rabbit-server.{port}.log"),
            "RABBITMQ_MNESIA_BASE": str(path / "mnesia"),
            "RABBITMQ_ENABLED_PLUGINS_FILE": str(plugin_path / "plugins"),
            "RABBITMQ_NODE_PORT": str(port),
            "RABBITMQ_DIST_PORT": str(distribution_port),
            # Use the port number in node name, so multiple instances started
            # at different ports will work separately instead of clustering.
            "RABBITMQ_NODENAME": node_name or f"rabbitmq-test-{port}",
        }
        super().__init__(command, host, port, timeout=60, envvars=envvars)
        self.rabbit_ctl = rabbit_ctl

    def rabbitctl_output(self, *args: str) -> str:
        """Query rabbitctl with args.

        :param list args: list of additional args to query
        """
        ctl_command: List[str] = [self.rabbit_ctl]
        ctl_command.extend(args)
        return subprocess.check_output(ctl_command, env=self._popen_kwargs["env"]).decode("utf-8")

    def list_exchanges(self) -> List[str]:
        """Get exchanges defined on given rabbitmq."""
        exchanges: List[str] = []
        output = self.rabbitctl_output("list_exchanges", "name")
        unwanted_exchanges = ["Listing exchanges for vhost / ...", "...done."]

        for exchange in output.split("\n"):
            if exchange and exchange not in unwanted_exchanges:
                exchanges.append(str(exchange))

        return exchanges

    def list_queues(self) -> List[str]:
        """Get queues defined on given rabbitmq."""
        queues: List[str] = []
        output = self.rabbitctl_output("list_queues", "name")

        for queue in output.split("\n"):
            if queue and not self._UNWANTED_QUEUE_PATTERN.search(queue.strip(". ").lower()):
                queues.append(str(queue))

        return queues



# Copyright (C) 2013 by Clearcode <http://clearcode.cc>
# and associates (see AUTHORS).

# This file is part of pytest-rabbitmq.

# pytest-rabbitmq is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pytest-rabbitmq is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with pytest-rabbitmq.  If not, see <http://www.gnu.org/licenses/>.
"""RabbitMQ client fixture factory."""
import logging
from typing import Callable, Generator

import pytest
from pika import BlockingConnection, ConnectionParameters
from pika.credentials import PlainCredentials
from pika.exceptions import ChannelClosed
from pytest import FixtureRequest

from pytest_rabbitmq.factories.executor import RabbitMqExecutor

logger = logging.getLogger("pytest-rabbitmq")


def clear_rabbitmq(process: RabbitMqExecutor, rabbitmq_connection: BlockingConnection) -> None:
    """Clear queues and exchanges from given rabbitmq process.

    :param RabbitMqExecutor process: rabbitmq process
    :param pika.connection.Connection rabbitmq_connection: connection to rabbitmq

    """
    channel = rabbitmq_connection.channel()

    for exchange in process.list_exchanges():
        if exchange.startswith("amq."):
            # ----------------------------------------------------------------
            # From rabbit docs:
            # https://www.rabbitmq.com/amqp-0-9-1-reference.html
            # ----------------------------------------------------------------
            # Exchange names starting with "amq." are reserved for pre-declared
            # and standardised exchanges. The client MAY declare an exchange
            # starting with "amq." if the passive option is set, or the
            # exchange already exists. Error code: access-refused
            # ----------------------------------------------------------------
            continue
        channel.exchange_delete(exchange)

    for queue_name in process.list_queues():
        if queue_name.startswith("amq."):
            # ----------------------------------------------------------------
            # From rabbit docs:
            # https://www.rabbitmq.com/amqp-0-9-1-reference.html
            # ----------------------------------------------------------------
            # Queue names starting with "amq." are reserved for pre-declared
            # and standardised queues. The client MAY declare a queue starting
            # with "amq." if the passive option is set, or the queue already
            # exists. Error code: access-refused
            # ----------------------------------------------------------------
            continue
        channel.queue_delete(queue_name)


def rabbitmq(
    process_fixture_name: str,
    teardown: Callable[[RabbitMqExecutor, BlockingConnection], None] = clear_rabbitmq,
) -> Callable[[FixtureRequest], Generator[BlockingConnection, None, None]]:
    """Client fixture factory for RabbitMQ.

    :param str process_fixture_name: name of RabbitMQ process variable
        returned by rabbitmq_proc
    :param callable teardown: custom callable that clears rabbitmq

    .. note::

        calls to rabbitmqctl might be as slow or even slower
        as restarting process. To speed up, provide Your own teardown function,
        to remove queues and exchanges of your choosing, without querying
        rabbitmqctl underneath.

    :returns RabbitMQ connection
    """

    @pytest.fixture
    def rabbitmq_factory(request: FixtureRequest) -> Generator[BlockingConnection, None, None]:
        """Client fixture for RabbitMQ.

        #. Get module and config.
        #. Connect to RabbitMQ using the parameters from config.

        :param TCPExecutor rabbitmq_proc: tcp executor
        :param FixtureRequest request: fixture request object
        :rtype: pika.adapters.blocking_connection.BlockingConnection
        :returns: instance of :class:`BlockingConnection`
        """
        # load required process fixture
        process = request.getfixturevalue(process_fixture_name)

        credentials = PlainCredentials("guest", "guest")
        parameters = ConnectionParameters(
            host=process.host, port=process.port, virtual_host="/", credentials=credentials
        )
        connection = BlockingConnection(parameters)

        yield connection
        teardown(process, connection)
        try:
            connection.close()
        except ChannelClosed as e:
            # at this stage this exception occurs when connection is being closed
            logger.warning(f"ChannelClosedException occured while closing connection {e}")

    return rabbitmq_factory



# Copyright (C) 2013 by Clearcode <http://clearcode.cc>
# and associates (see AUTHORS).

# This file is part of pytest-rabbitmq.

# pytest-rabbitmq is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pytest-rabbitmq is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with pytest-rabbitmq.  If not, see <http://www.gnu.org/licenses/>.
"""Plugin definition for pytest-rabbitmq."""
from tempfile import gettempdir

from pytest import Parser

from pytest_rabbitmq import factories

# pylint:disable=invalid-name
_help_ctl = "RabbitMQ ctl path"
_help_server = "RabbitMQ server path"
_help_logsdir = "Logs directory location"
_help_plugindir = "Directory where 'plugin' file is located"
_help_host = "Host at which RabbitMQ will accept connections"
_help_port = "Port at which RabbitMQ will accept connections"
_help_distribution_port = "Port at which RabbitMQ nodes will communicate with each other"
_help_node = "Node name for rabbitmq instance"


def pytest_addoption(parser: Parser) -> None:
    """Confioguration option."""
    parser.addini(name="rabbitmq_host", help=_help_host, default="127.0.0.1")
    parser.addini(
        name="rabbitmq_port",
        help=_help_port,
        default=None,
    )
    parser.addini(
        name="rabbitmq_distribution_port",
        help=_help_distribution_port,
        default=None,
    )
    parser.addini(
        name="rabbitmq_ctl",
        help=_help_ctl,
        default="/usr/lib/rabbitmq/bin/rabbitmqctl",
    )
    parser.addini(
        name="rabbitmq_server",
        help=_help_server,
        default="/usr/lib/rabbitmq/bin/rabbitmq-server",
    )
    parser.addini(
        name="rabbitmq_logsdir",
        help=_help_logsdir,
        default=None,
    )
    parser.addini(
        name="rabbitmq_plugindir",
        help=_help_logsdir,
        default=gettempdir(),
    )
    parser.addini(
        name="rabbitmq_node",
        help=_help_node,
        default=None,
    )

    parser.addoption(
        "--rabbitmq-host",
        action="store",
        dest="rabbitmq_host",
        help=_help_host,
    )
    parser.addoption("--rabbitmq-port", action="store", dest="rabbitmq_port", help=_help_port)
    parser.addoption(
        "--rabbitmq-distribution-port",
        action="store",
        dest="rabbitmq_distribution_port",
        help=_help_distribution_port,
    )
    parser.addoption("--rabbitmq-ctl", action="store", dest="rabbitmq_ctl", help=_help_ctl)
    parser.addoption("--rabbitmq-server", action="store", dest="rabbitmq_server", help=_help_server)
    parser.addoption(
        "--rabbitmq-logsdir",
        action="store",
        metavar="path",
        dest="rabbitmq_logsdir",
        help=_help_logsdir,
    )
    parser.addoption(
        "--rabbitmq-plugindir",
        action="store",
        metavar="path",
        dest="rabbitmq_plugindir",
        help=_help_plugindir,
    )
    parser.addoption(
        "--rabbitmq-node",
        action="store",
        dest="rabbitmq_node",
        help=_help_node,
    )


rabbitmq_proc = factories.rabbitmq_proc()
rabbitmq = factories.rabbitmq("rabbitmq_proc")



# Copyright (C) 2014 by Clearcode <http://clearcode.cc>
# and associates (see AUTHORS).

# This file is part of pytest-rabbitmq.

# pytest-rabbitmq is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pytest-rabbitmq is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with pytest-rabbitmq.  If not, see <http://www.gnu.org/licenses/>.
"""RabbitMQ process fixture factory."""

from pathlib import Path
from typing import (
    Any,
    Callable,
    Generator,
    List,
    Optional,
    Set,
    Tuple,
    TypedDict,
    Union,
)
from warnings import warn

import pytest
from mirakuru.exceptions import ProcessExitedWithError
from port_for import get_port
from pytest import FixtureRequest, TempPathFactory

from pytest_rabbitmq.factories.executor import RabbitMqExecutor

PortType = Union[
    None,
    str,
    int,
    Tuple[int, int],
    Set[int],
    List[str],
    List[int],
    List[Tuple[int, int]],
    List[Set[int]],
    List[Union[Set[int], Tuple[int, int]]],
    List[Union[str, int, Tuple[int, int], Set[int]]],
]


class RabbitMQConfig(TypedDict):
    """Pytest RabbitMQ config definition type."""

    host: str
    port: PortType
    distribution_port: PortType
    logsdir: Optional[Path]
    server: str
    ctl: str
    node: str
    plugindir: Path


def get_config(request: FixtureRequest) -> RabbitMQConfig:
    """Return a dictionary with config options."""

    def get_conf_option(option: str) -> Any:
        option_name = "rabbitmq_" + option
        return request.config.getoption(option_name) or request.config.getini(option_name)

    port = get_conf_option("port")
    distribution_port = get_conf_option("distribution_port")
    logsdir = get_conf_option("logsdir")
    config: RabbitMQConfig = {
        "host": get_conf_option("host"),
        "port": int(port) if port else None,
        "distribution_port": int(distribution_port) if distribution_port else None,
        "logsdir": Path(logsdir) if logsdir else None,
        "server": get_conf_option("server"),
        "ctl": get_conf_option("ctl"),
        "node": get_conf_option("node"),
        "plugindir": Path(get_conf_option("plugindir")),
    }
    return config


def rabbitmq_proc(
    server: Optional[str] = None,
    host: Optional[str] = None,
    port: PortType = -1,
    distribution_port: PortType = -1,
    node: Optional[str] = None,
    ctl: Optional[str] = None,
    logsdir: Optional[Path] = None,
    plugindir: Optional[Path] = None,
) -> Callable[[FixtureRequest, TempPathFactory], Generator[RabbitMqExecutor, None, None]]:
    """Fixture factory for RabbitMQ process.

    :param server: path to rabbitmq-server command
    :param host: server host
    :param port:
        exact port (e.g. '8000', 8000)
        randomly selected port (None) - any random available port
        [(2000,3000)] or (2000,3000) - random available port from a given range
        [{4002,4003}] or {4002,4003} - random of 4002 or 4003 ports
        [(2000,3000), {4002,4003}] -random of given range and set
    :param distribution_port:
        exact port (e.g. '8000', 8000)
        randomly selected port (None) - any random available port
        [(2000,3000)] or (2000,3000) - random available port from a given range
        [{4002,4003}] or {4002,4003} - random of 4002 or 4003 ports
        [(2000,3000), {4002,4003}] -random of given range and set
    :param node: RabbitMQ node name used for setting environment
                          variable RABBITMQ_NODENAME (the default depends
                          on the port number, so multiple nodes are not
                          clustered)
    :param ctl: path to rabbitmqctl file
    :param logsdir: path to log directory

    :returns pytest fixture with RabbitMQ process executor
    """

    @pytest.fixture(scope="session")
    def rabbitmq_proc_fixture(
        request: FixtureRequest, tmp_path_factory: TempPathFactory
    ) -> Generator[RabbitMqExecutor, None, None]:
        """Fixture for RabbitMQ process.

        #. Get config.
        #. Make a temporary directory.
        #. Setup required environment variables:
        #.  * RABBITMQ_LOG_BASE
        #.  * RABBITMQ_MNESIA_BASE
        #.  * RABBITMQ_ENABLED_PLUGINS_FILE
        #.  * RABBITMQ_NODE_PORT
        #.  * RABBITMQ_DIST_PORT
        #.  * RABBITMQ_NODENAME
        #. Start a rabbit server
            `<http://www.rabbitmq.com/man/rabbitmq-server.1.man.html>`_
        #. Stop rabbit server and remove temporary files after tests.

        :param FixtureRequest request: fixture request object
        :rtype: pytest_rabbitmq.executors.TCPExecutor
        :returns: tcp executor of running rabbitmq-server
        """
        config = get_config(request)
        rabbit_ctl = ctl or config["ctl"]
        rabbit_server = server or config["server"]
        rabbit_host = host or config["host"]
        rabbit_port = get_port(port) or get_port(config["port"])
        assert rabbit_port
        rabbit_distribution_port = get_port(distribution_port, [rabbit_port]) or get_port(
            config["distribution_port"], [rabbit_port]
        )
        assert rabbit_distribution_port
        assert (
            rabbit_distribution_port != rabbit_port
        ), "rabbit_port and distribution_port can not be the same!"

        tmpdir = tmp_path_factory.mktemp(f"pytest-rabbitmq-{request.fixturename}")

        rabbit_plugin_path = plugindir or config["plugindir"]

        rabbit_logpath = config["logsdir"] or logsdir
        if rabbit_logpath:
            warn(
                f"rabbitmq_logsdir and --rabbitmq-logsdir config option is "
                f"deprecated, and will be dropped in future releases. "
                f"All fixture related data resides within {tmpdir}",
                DeprecationWarning,
            )
        if not rabbit_logpath:
            rabbit_logpath = tmpdir / "logs"

        rabbit_executor = RabbitMqExecutor(
            rabbit_server,
            rabbit_host,
            rabbit_port,
            rabbit_distribution_port,
            rabbit_ctl,
            logpath=rabbit_logpath,
            path=tmpdir,
            plugin_path=rabbit_plugin_path,
            node_name=node or config["node"],
        )

        rabbit_executor.start()
        yield rabbit_executor
        try:
            rabbit_executor.stop()
        except ProcessExitedWithError:
            pass

    return rabbitmq_proc_fixture
